-------------------------------------------------------------------------------
--! @file       CipherCore.vhd
--! @brief      Cipher core for CLOC V2
--! @project    CAESAR Candidate Evaluation
--! @author     
--! @copyright  
--!             
--!             
--! @license    
--!             
--!             
--!             
--! @note       
--!             
--!             
-------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.ALL;
use ieee.numeric_std.ALL;
use work.AEAD_pkg.ALL;
use work.AES_pkg.ALL;

use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;



entity CipherCore is
    generic (
        --! Reset behavior
        G_ASYNC_RSTN    : boolean := False; --! Async active low reset
        --! Block size (bits)
        G_DBLK_SIZE     : integer := 128;    --! Data   Block size n(bits) : 128(AES128)/64(TWINE80)
        G_KEY_SIZE      : integer := 128;    --! Key    Key length(bits)   : 128(AES128)/80(TWINE80)
        G_TAG_SIZE      : integer := 64;     --! Tag
        --! The number of bits required to hold block size expressed in
        --! bytes = log2_ceil(G_DBLK_SIZE/8)
        G_LBS_BYTES     : integer := 4;
        G_MAX_LEN       : integer := SINGLE_PASS_MAX;        --(+)caesar_hw_devpkg-1.0-3

--v2(+) from
        G_PARAM         : std_logic_vector(7 downto 0)    := X"C0"    --param (AES-128, Nonce=12bytes, Tag= 8bytes)
--        G_PARAM       : std_logic_vector(7 downto 0)    := X"D0"    --param (AES-128, Nonce= 8bytes, Tag= 8bytes)
--v2(+) to
    );

    port (
        --! Global
        clk             : in  std_logic;
        rst             : in  std_logic;
        --! PreProcessor (data)
        key             : in  std_logic_vector(G_KEY_SIZE        -1 downto 0);
        bdi             : in  std_logic_vector(G_DBLK_SIZE       -1 downto 0);
        --! PreProcessor (controls)
        key_ready       : out std_logic;
        key_valid       : in  std_logic;
        key_update      : in  std_logic;
        decrypt         : in  std_logic;        -- '0' : ENC / '1' : DEC
        bdi_ready       : out std_logic;
        bdi_valid       : in  std_logic;
        bdi_type        : in  std_logic_vector(3                -1 downto 0);
        bdi_partial     : in  std_logic;
        bdi_eot         : in  std_logic;
        bdi_eoi         : in  std_logic;
        bdi_size        : in  std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        bdi_valid_bytes : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        bdi_pad_loc     : in  std_logic_vector(G_DBLK_SIZE/8    -1 downto 0);
        --! PostProcessor
        bdo             : out std_logic_vector(G_DBLK_SIZE      -1 downto 0);
        bdo_valid       : out std_logic;
        bdo_ready       : in  std_logic;
        bdo_size        : out std_logic_vector(G_LBS_BYTES+1    -1 downto 0);
        msg_auth_done   : out std_logic;
        msg_auth_valid  : out std_logic
    );

end entity CipherCore;

architecture structure of CipherCore is

-------------------------------------------------------
-- components
-------------------------------------------------------
    -- AES128(GMU Source)
    component AES_Enc_KOF is
        port(
            clk         : in  std_logic;
            rst         : in  std_logic;
            din         : in  std_logic_vector(G_DBLK_SIZE-1 downto 0);
            key         : in  std_logic_vector(AES_BLOCK_SIZE-1 downto 0);
            dout        : out std_logic_vector(AES_BLOCK_SIZE-1 downto 0);

            start       : in  std_logic;
            ready       : out std_logic;
            done        : out std_logic
        );
    end component;



-------------------------------------------------------
-- constants
-------------------------------------------------------

    constant    C_BIT_SIZE      : std_logic_vector(7 downto 0)              := CONV_STD_LOGIC_VECTOR(G_DBLK_SIZE, 8);    -- bit width
    constant    C_TAG_MASK_0    : std_logic_vector((G_DBLK_SIZE-G_TAG_SIZE)-1 downto 0) := CONV_STD_LOGIC_VECTOR(0,(G_DBLK_SIZE-G_TAG_SIZE));    -- ALL'0'
    constant    C_TAG_MASK_1    : std_logic_vector(G_TAG_SIZE-1 downto 0)   := CONV_STD_LOGIC_VECTOR((2**G_TAG_SIZE-1),     G_TAG_SIZE);        -- ALL'1'
    constant    C_TAG_MASK      : std_logic_vector(G_DBLK_SIZE-1 downto 0)  := C_TAG_MASK_1 & C_TAG_MASK_0;

-------------------------------------------------------
-- signals
-------------------------------------------------------
    type    m_state is (    
                            ST_INIT             ,
                            ST_CHECK_KEY        ,
                            ST_ACTKEY           ,
                            ST_LD_N             ,
                            ST_WAIT_FIRST_A     ,
                            ST_WAIT_AES_A       ,
                            ST_WAIT_AI          ,
                            ST_WAIT_AES_LASTA   ,
                            ST_USE_N            ,
                            ST_SETINP_N         ,
                            ST_WAIT_AES0_N      ,
                            ST_WAIT_AES1_N      ,
                            ST_WAIT_M           ,
                            ST_WAIT_AES0_M      ,
                            ST_WAIT_AES1_M      ,
                            ST_SETINP1A_LASTM   ,
                            ST_SETINP1B_LASTM   ,
                            ST_WAIT_AES1_LASTM  ,
                            ST_LAST1            ,
                            ST_LAST2            ,
                            ST_WAIT_AES2_A      ,
                            ST_WAIT_AES2_LASTM  
                        );



    signal    state_reg             : m_state;

    signal    aes_in0_reg           : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    signal    aes_in1_reg           : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    signal    aes_out0_reg          : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    signal    aes_out1_reg          : std_logic_vector(G_DBLK_SIZE-1 downto 0);

    signal    aes_in_node           : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    signal    aes_out_node          : std_logic_vector(G_DBLK_SIZE-1 downto 0);

    signal    aes_outsel_reg        : std_logic;
    signal    aes_start_reg         : std_logic;
    signal    aes_a1_msb_reg        : std_logic;
    signal    aes_last_len_reg      : std_logic_vector(7 downto 0);
    signal    aes_finish            : std_logic;

    signal    ozp_in0_node          : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    signal    ozp_in1_node          : std_logic_vector(7 downto 0);
    signal    ozp_in2_node          : std_logic;
    signal    ozp_out_node          : std_logic_vector(G_DBLK_SIZE-1 downto 0);

    signal    f1_in0_node           : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    signal    f2_in0_node           : std_logic_vector(G_DBLK_SIZE-1 downto 0);

    -- for CAESAR HW API
    signal    d_vld                 : std_logic;
    signal    t_vld                 : std_logic;

    signal    tag_out               : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    signal    bdo_out0_node         : std_logic_vector(G_DBLK_SIZE-1 downto 0);

    signal    nonce_reg             : std_logic_vector(G_DBLK_SIZE-1 downto 0);
    signal    nonce_size_reg        : std_logic_vector(7 downto 0);        -- Nonce Length
    signal    key_reg               : std_logic_vector(G_KEY_SIZE-1 downto 0);
    signal    key_ready_reg         : std_logic;
    signal    decrypt_reg           : std_logic;
    signal    bdi_ready_reg         : std_logic;
    signal    bdi_size_reg          : std_logic_vector(G_LBS_BYTES+1-1 downto 0);
    signal    last_seg              : std_logic;
    signal    last_len              : std_logic_vector(7 downto 0);        -- by bits

    signal    bdi_eoi_reg           : std_logic;
    signal    input_flg_a           : std_logic;                        
    signal    input_flg_d           : std_logic;                        
    signal    zero_len_flg_a        : std_logic;                        
    signal    zero_len_flg_d        : std_logic;                        
    signal    proc_flg_a            : std_logic;                        
    signal    proc_flg_d            : std_logic;                        



-------------------------------------------------------
-- functions
-------------------------------------------------------
    -- ozp (one zero padding (1000...))
    function ozp_cut_func 
                (    a      : std_logic_vector(G_DBLK_SIZE-1 downto 0);
                    len     : std_logic_vector(7 downto 0);
                    ozp     : std_logic            -- padding mode (0 : 000..000 padding, 1 : 100...000 padding)
                )
                        return std_logic_vector is

        type    maskarray is array(a'range) of std_logic_vector(a'range);

        variable    and_mask    : maskarray;
        variable    or_mask     : maskarray;

        variable    am          : std_logic_vector(a'range);
        variable    om          : std_logic_vector(a'range);
        variable    am_sel      : std_logic_vector(a'range);

        variable    selv        : integer;

        variable    y           : std_logic_vector(a'range);

        begin

            selv    := CONV_INTEGER(len(6 downto 0));


            if (len = C_BIT_SIZE) then
                y    := a;

            else
                --  len        AND_mask        OR_mask
                --    0        0000...0000     1000...0000
                --    1        1000...0000     0100...0000
                --    2        1100...0000     0010...0000
                --  ...
                --  126        1111...1100     0000...0010
                --  127        1111...1110     0000...0001
                and_mask(0)     := (others=>'0');
                or_mask(0)      := '1' & CONV_STD_LOGIC_VECTOR(0, G_DBLK_SIZE-1);
                for I in 1 to G_DBLK_SIZE-1 loop
                    am                := and_mask(I - 1);
                    and_mask(I)       := '1' & am(G_DBLK_SIZE-1 downto 1);

                    om                := or_mask(I - 1);
                    or_mask(I)        := '0' & om(G_DBLK_SIZE-1 downto 1);
                end loop;

                am_sel   := a and and_mask(selv);

                if (ozp = '1') then
                    y    := am_sel or or_mask(selv);
                else
                    y    := am_sel;
                end if;

            end if;

            return y;
    end;



    -- fix0 (MSBit 0 fixed)
    function fix0_func 
                (    a      : std_logic_vector(G_DBLK_SIZE-1 downto 0)
                )
                        return std_logic_vector is

        variable    y       : std_logic_vector(a'range);

        begin
            y    := '0' & a(G_DBLK_SIZE-2 downto 0);
            return y;
    end;

    -- fix1 (MSBit 1 fixed)
    function fix1_func 
                (    a      : std_logic_vector(G_DBLK_SIZE-1 downto 0)
                )
                        return std_logic_vector is

        variable    y       : std_logic_vector(a'range);

        begin
            y    := '1' & a(G_DBLK_SIZE-2 downto 0);
            return y;
    end;



    -- h
    function h_func 
                (    a            : std_logic_vector(G_DBLK_SIZE-1 downto 0)
                )
                        return std_logic_vector is

        variable    w3            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w2            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w1            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w0            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);

        variable    y             : std_logic_vector(a'range);

        begin

            w3    := a(G_DBLK_SIZE-1         downto (G_DBLK_SIZE/4)*3);     -- X[1]
            w2    := a((G_DBLK_SIZE/4)*3-1     downto (G_DBLK_SIZE/4)*2);   -- X[2]
            w1    := a((G_DBLK_SIZE/4)*2-1     downto (G_DBLK_SIZE/4));     -- X[3]
            w0    := a((G_DBLK_SIZE/4)-1    downto 0);                      -- X[4]

            -- X[1,2], X[2,3], X[3,4], X[1,2,4]
            y     := (w3 xor w2) & (w2 xor w1) & (w1 xor w0) & (w3 xor w2 xor w0);
            return y;
    end;

    -- f1
    function f1_func 
                (    a            : std_logic_vector(G_DBLK_SIZE-1 downto 0)
                )
                        return std_logic_vector is

        variable    w3            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w2            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w1            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w0            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);

        variable    y             : std_logic_vector(a'range);

        begin

            w3    := a(G_DBLK_SIZE-1         downto (G_DBLK_SIZE/4)*3);     -- X[1]
            w2    := a((G_DBLK_SIZE/4)*3-1     downto (G_DBLK_SIZE/4)*2);   -- X[2]
            w1    := a((G_DBLK_SIZE/4)*2-1     downto (G_DBLK_SIZE/4));     -- X[3]
            w0    := a((G_DBLK_SIZE/4)-1    downto 0);                      -- X[4]

            -- X[1,3], X[2,4], X[1,2,3], X[2,3,4]
            y     := (w3 xor w1) & (w2 xor w0) & (w3 xor w2 xor w1) & (w2 xor w1 xor w0);
            return y;
    end;

    -- f2
    function f2_func 
                (    a            : std_logic_vector(G_DBLK_SIZE-1 downto 0)
                )
                        return std_logic_vector is

        variable    w3            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w2            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w1            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w0            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);

        variable    y             : std_logic_vector(a'range);

        begin

            w3    := a(G_DBLK_SIZE-1         downto (G_DBLK_SIZE/4)*3);     -- X[1]
            w2    := a((G_DBLK_SIZE/4)*3-1     downto (G_DBLK_SIZE/4)*2);   -- X[2]
            w1    := a((G_DBLK_SIZE/4)*2-1     downto (G_DBLK_SIZE/4));     -- X[3]
            w0    := a((G_DBLK_SIZE/4)-1    downto 0);                      -- X[4]

            -- X[2], X[3], X[4], X[1,2]
            y     := (w2) & (w1) & (w0) & (w3 xor w2);
            return y;
    end;

    -- g1
    function g1_func 
                (    a            : std_logic_vector(G_DBLK_SIZE-1 downto 0)
                )
                        return std_logic_vector is

        variable    w3            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w2            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w1            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);
        variable    w0            : std_logic_vector(G_DBLK_SIZE/4-1 downto 0);

        variable    y             : std_logic_vector(a'range);

        begin

            w3    := a(G_DBLK_SIZE-1         downto (G_DBLK_SIZE/4)*3);     -- X[1]
            w2    := a((G_DBLK_SIZE/4)*3-1     downto (G_DBLK_SIZE/4)*2);   -- X[2]
            w1    := a((G_DBLK_SIZE/4)*2-1     downto (G_DBLK_SIZE/4));     -- X[3]
            w0    := a((G_DBLK_SIZE/4)-1    downto 0);                      -- X[4]

            -- X[3], X[4], X[1,2], X[2,3]
            y     := (w1) & (w0) & (w3 xor w2) & (w2 xor w1);
            return y;
    end;



-------------------------------------------------------------------------------

begin



-------------------------------------------------------
-- ENCRYPT CORE
-------------------------------------------------------
    aes_in_node    <= fix1_func(a=>aes_in0_reg)    when (state_reg = ST_WAIT_AES0_M) else    aes_in0_reg;

    u0 : AES_Enc_KOF
        port map(
            clk         => clk,
            rst         => rst,
            din         => aes_in_node,
            key         => key_reg,
            dout        => aes_out_node,

            start       => aes_start_reg,
            ready       => open,
            done        => aes_finish
    );



-------------------------------------------------------
-- A,D input, last data
-------------------------------------------------------
    process(clk, rst)
    begin
        if (rst = '1') then
            
            bdi_eoi_reg         <= '0';
            input_flg_a         <= '0';
            input_flg_d         <= '0';
            bdi_size_reg        <= (others=>'0');

        elsif (clk'event and clk = '1') then
            if (state_reg = ST_INIT) then
                bdi_eoi_reg        <= '0';
                input_flg_a        <= '0';
                input_flg_d        <= '0';

            elsif ( bdi_valid    = '1') then
                if (bdi_eoi = '1') then
                    bdi_eoi_reg    <= '1';
                end if;

                if (bdi_type(2 downto 1) = ST_A) then
                    input_flg_a    <= '1';
                end if;

                if (bdi_type(2 downto 1) = ST_D) then
                    input_flg_d    <= '1';
                end if;

                bdi_size_reg    <= bdi_size;
            end if;
        end if;
    end process;

    zero_len_flg_a    <= bdi_eoi_reg and (not input_flg_a);
    zero_len_flg_d    <= bdi_eoi_reg and (not input_flg_d);


    last_seg    <= '1'                when (    
                                            (proc_flg_a    = '1'    and zero_len_flg_a = '1')        or
                                            (proc_flg_d    = '1'    and zero_len_flg_d = '1')
                                    ) else
                    bdi_eot;

    last_len    <=     X"00"            when (    
                                            (proc_flg_a    = '1'    and zero_len_flg_a = '1')        or
                                            (proc_flg_d    = '1'    and zero_len_flg_d = '1')
                                    ) else
                    bdi_size &"000";



-------------------------------------------------------
-- main contol
-------------------------------------------------------
    -- ozp_cut_func
    ozp_in0_node    <=    aes_out_node      when (state_reg = ST_WAIT_AES1_LASTM) else
                        aes_out0_reg        when (        state_reg = ST_WAIT_M                or
                                                    (    state_reg = ST_SETINP1A_LASTM    and
                                                        decrypt_reg = '0'
                                                    )
                                                ) else

--v2(-)
--                        nonce_reg            when (state_reg = ST_USE_N) else            -- SH[1] xor ozp(N)
--v2(+) from
                        G_PARAM & nonce_reg(G_DBLK_SIZE-1 downto 8)    when (state_reg = ST_USE_N) else            -- SH[1] xor ozp(N)
--v2(+) to
                        bdi;


    ozp_in1_node    <=    aes_last_len_reg    when (    state_reg = ST_WAIT_AES1_LASTM    or
                                                    state_reg = ST_WAIT_AES2_LASTM
                                                ) else

--v2(-)
--                        nonce_size_reg        when (state_reg = ST_USE_N) else
--v2(+) from
                        (nonce_size_reg + X"08")    when (state_reg = ST_USE_N) else
--v2(+) to
                        last_len;


    ozp_in2_node    <=    '0'           when (    state_reg = ST_WAIT_M             or
                                                  state_reg = ST_WAIT_AES1_LASTM    or
                                                  state_reg = ST_WAIT_AES2_LASTM
                                                ) else
                        '1';

    ozp_out_node    <=    ozp_cut_func(a=>ozp_in0_node, len=>ozp_in1_node, ozp=>ozp_in2_node);


    bdo_out0_node    <= bdi xor aes_out0_reg;
    f1_in0_node        <= f1_func(a=>aes_in0_reg);
    f2_in0_node        <= f2_func(a=>aes_in0_reg);



    process(clk, rst)
    begin
        if (rst = '1') then
            aes_in0_reg         <= (others=>'0');
            aes_out0_reg        <= (others=>'0');
            aes_out1_reg        <= (others=>'0');
            aes_in1_reg         <= (others=>'0');
            aes_outsel_reg      <= '0';
            aes_start_reg       <= '0';
            aes_a1_msb_reg      <= '0';
            aes_last_len_reg    <= X"00";

            bdi_ready_reg       <= '0';
            key_ready_reg       <= '0';
            key_reg             <= (others=>'0');

            decrypt_reg         <= '0';
            nonce_reg           <= (others=>'0');
            nonce_size_reg      <= (others=>'0');
            proc_flg_a          <= '0';
            proc_flg_d          <= '0';

            state_reg           <= ST_INIT;

        elsif (clk'event and clk = '1') then

            case (state_reg) is


                when ST_INIT                =>
                    aes_outsel_reg      <= '0';
                    aes_start_reg       <= '0';

                    key_ready_reg       <= '0';

                    proc_flg_a          <= '0';
                    proc_flg_d          <= '0';

                    if (bdi_valid = '1'    and
                        bdi_type    = BDI_TYPE_NPUB
                    ) then
                        nonce_reg         <= bdi;
                        nonce_size_reg    <= last_len;

                        state_reg    <= ST_CHECK_KEY;
                    end if;
                    bdi_ready_reg        <= '0';


                when ST_CHECK_KEY            =>             
                    if (key_update = '1') then              
                        bdi_ready_reg    <= '0';
                        state_reg        <= ST_ACTKEY;

                    else                                
                        bdi_ready_reg    <= '1';            
                        state_reg        <= ST_LD_N;
                    end if;

                when ST_ACTKEY                =>            
                    if (key_valid = '1') then
                        key_ready_reg       <= '1';
                        key_reg             <= key;

                        bdi_ready_reg    <= '1';            
                        state_reg        <= ST_LD_N;
                    end if;

                when ST_LD_N                =>              
                    key_ready_reg     <= '0';
                    bdi_ready_reg     <= '0';

                    decrypt_reg       <= decrypt;
                    proc_flg_a        <= '1';
                    proc_flg_d        <= '0';

                    state_reg    <= ST_WAIT_FIRST_A;


                when ST_WAIT_FIRST_A        =>        -- wait A[1]
                    if (    (   bdi_valid = '1'    and
                                bdi_type(2 downto 1) = ST_A)    or      
                            zero_len_flg_a = '1'                    
                    ) then
                        bdi_ready_reg       <= not zero_len_flg_a;

                        aes_in0_reg         <= fix0_func(a=>ozp_out_node);
                        aes_last_len_reg    <= last_len;

                        aes_start_reg       <= '1';

                        if (last_seg = '1') then    
                            aes_a1_msb_reg   <= ozp_out_node(G_DBLK_SIZE-1);    -- MSB of ozp(A[1])
                            state_reg        <= ST_WAIT_AES2_A;
                        else                        
                            aes_a1_msb_reg   <= bdi(G_DBLK_SIZE-1);            -- MSB of A[1]
                            state_reg        <= ST_WAIT_AES_A;
                        end if;

                    else
                        bdi_ready_reg        <= '0';
                    end if;


                -------------------------------------------------------
                -- A[i]
                -------------------------------------------------------
                when ST_WAIT_AES_A    =>        -- wait AES computation
                    bdi_ready_reg    <= '0';
                    aes_start_reg    <= '0';

                    if (aes_finish = '1') then
                        if (aes_a1_msb_reg = '1') then
                            aes_out0_reg    <= h_func(a=>aes_out_node);
                        else
                            aes_out0_reg    <= aes_out_node;
                        end if;

                        state_reg        <= ST_WAIT_AI;

                    end if;

                when ST_WAIT_AI        =>        -- 
                    aes_a1_msb_reg        <= '0';

                    if (bdi_valid = '1') then
                        aes_start_reg        <= '1';
                        bdi_ready_reg        <= '1';

                        if (last_seg = '1') then
                            aes_in0_reg            <= ozp_out_node xor aes_out0_reg;
                            aes_last_len_reg     <= last_len;
                            state_reg            <= ST_WAIT_AES_LASTA;
                        else
                            aes_in0_reg          <= bdo_out0_node;
                            state_reg            <= ST_WAIT_AES_A;
                        end if;
                    end if;

                when ST_WAIT_AES_LASTA    =>    -- wait AES computation
                    bdi_ready_reg    <= '0';
                    aes_start_reg    <= '0';

                    if (aes_finish = '1') then
                        aes_out0_reg    <= aes_out_node;
                        state_reg       <= ST_USE_N;
                    end if;

                when ST_WAIT_AES2_A        =>    -- wait AES computation
                    bdi_ready_reg       <= '0';
                    aes_start_reg       <= '0';

                    if (aes_finish = '1') then
                        if (aes_a1_msb_reg = '1') then
                            aes_out0_reg    <= h_func(a=>aes_out_node);
                        else
                            aes_out0_reg    <= aes_out_node;
                        end if;

                        state_reg        <= ST_USE_N;
                    end if;



                -------------------------------------------------------
                -- N
                -------------------------------------------------------
                when ST_USE_N            =>
                    aes_in0_reg     <= ozp_out_node xor aes_out0_reg;

                    state_reg       <= ST_SETINP_N;

                when ST_SETINP_N        =>
                    -- HASH f1,f2
                    -- f1,2_in0_node : f1_func(a=>aes_in0_reg)
                    if (aes_last_len_reg >= C_BIT_SIZE) then
                        aes_in0_reg        <= f1_in0_node;
                    else
                        aes_in0_reg        <= f2_in0_node;
                    end if;
                    aes_start_reg    <= '1';
                    state_reg        <= ST_WAIT_AES0_N;

                when ST_WAIT_AES0_N        =>    -- wait AES computation
                    if (aes_finish = '1') then
                        aes_out0_reg    <= aes_out_node;
                        aes_in0_reg     <= f2_in0_node;            -- g2_func = f2_func (Equivalent))
                        aes_in1_reg     <= g1_func(a=>aes_in0_reg);
                        aes_start_reg   <= '1';
                        state_reg       <= ST_WAIT_AES1_N;
                    else
                        aes_start_reg   <= '0';
                    end if;

                when ST_WAIT_AES1_N        =>    -- wait AES1 computation
                    aes_start_reg    <= '0';

                    if (aes_finish = '1') then
                        aes_outsel_reg    <= '0';

                        aes_out1_reg      <= aes_out_node;

                        proc_flg_a        <= '0';
                        proc_flg_d        <= '1';

                        state_reg         <= ST_WAIT_M;

                    end if;



                -------------------------------------------------------
                -- M[i]
                -------------------------------------------------------
                when ST_WAIT_M            =>
                    if (zero_len_flg_d = '1') then  
                            bdi_ready_reg    <= '0';

                            aes_in0_reg      <= aes_in1_reg;
                            aes_start_reg    <= '1';
                            state_reg        <= ST_WAIT_AES2_LASTM;

                    else
                        if (bdi_valid = '1'    and  
                            bdo_ready = '1'            -- PostProcessor Ready
                        ) then
                            bdi_ready_reg    <= '1';

                            if (last_seg = '1') then
                                aes_out0_reg     <= bdi xor ozp_out_node;        -- C[M}
                                aes_start_reg    <= '0';
                                state_reg        <= ST_SETINP1A_LASTM;

                            else
                                if (decrypt_reg = '1') then
                                    aes_in0_reg     <= bdi;
                                else
                                    aes_in0_reg     <= bdo_out0_node;
                                end if;
                                aes_start_reg       <= '1';
                                state_reg           <= ST_WAIT_AES0_M;
                            end if;

                        else
                            bdi_ready_reg    <= '0';
                        end if;
                    end if;


                when ST_WAIT_AES0_M        =>    -- wait AES0 computation
                    bdi_ready_reg    <= '0';

                    if (aes_finish = '1') then
                        aes_out0_reg    <= aes_out_node;
                        aes_in0_reg     <= aes_in0_reg xor aes_out1_reg;
                        aes_start_reg   <= '1';
                        state_reg       <= ST_WAIT_AES1_M;
                    else
                        aes_start_reg   <= '0';
                    end if;

                when ST_WAIT_AES1_M        =>    -- wait AES1 computation
                    aes_start_reg    <= '0';

                    if (aes_finish = '1') then
                        aes_out1_reg    <= aes_out_node;
                        state_reg       <= ST_WAIT_M;
                    end if;



                -------------------------------------------------------
                -- M[last]
                -------------------------------------------------------
                when ST_SETINP1A_LASTM    =>
                    bdi_ready_reg       <= '0';
                    aes_last_len_reg    <= last_len;        -- |C[m]|

                    aes_in0_reg        <= aes_out1_reg xor ozp_out_node;        -- Sp[m-1] xor ozp(C[m])
                    state_reg          <= ST_SETINP1B_LASTM;

                when ST_SETINP1B_LASTM    =>
                    -- PRF f1,f2
                    -- f1,2_in0_node : f1_func(a=>aes_in0_reg)
                    if (aes_last_len_reg >= C_BIT_SIZE) then
                        aes_in0_reg     <= f1_in0_node;
                    else
                        aes_in0_reg     <= f2_in0_node;
                    end if;
                    aes_start_reg       <= '1';
                    state_reg           <= ST_WAIT_AES1_LASTM;

                when ST_WAIT_AES1_LASTM    =>    -- wait AES1 computation
                    aes_start_reg    <= '0';

                    if (aes_finish = '1') then
                        aes_out1_reg    <= aes_out_node;    --  & C_TAG_MASK;
                        state_reg       <= ST_LAST1;
                    end if;

                when ST_WAIT_AES2_LASTM    =>    -- wait AES1 computation
                    aes_start_reg    <= '0';

                    if (aes_finish = '1') then
                        aes_out1_reg    <= aes_out_node;    --  & C_TAG_MASK;
                        state_reg       <= ST_LAST1;
                    end if;

                when ST_LAST1            =>    -- wait Tag computation
                    aes_outsel_reg    <= not decrypt_reg;   

                    state_reg    <= ST_LAST2;

                when ST_LAST2            =>    -- wait Tag computation
                    proc_flg_a        <= '0';
                    proc_flg_d        <= '0';

                    if (decrypt_reg    = '0') then  
                        bdi_ready_reg     <= '0';
                        aes_outsel_reg    <= '0';

                        if (bdo_ready    = '1') then
                            state_reg        <= ST_INIT;
                        end if;

                    else
                        if (bdi_valid = '1') then
                            bdi_ready_reg    <= '1';

                            state_reg        <= ST_INIT;
                        else
                            bdi_ready_reg    <= '0';
                        end if;

                    end if;


                when others        =>
                    state_reg        <= ST_INIT;
            end case;

        end if;
    end process;



-------------------------------------------------------
-- DO
-------------------------------------------------------
    tag_out        <=  (aes_out1_reg and C_TAG_MASK);

    d_vld        <= '1'        when (   state_reg       = ST_WAIT_M     and
                                        bdo_ready       = '1'           and
                                        bdi_valid       = '1'           and
                                        zero_len_flg_d  = '0'
                            ) else '0';

    t_vld        <= '1'        when (   state_reg       = ST_LAST2      and
                                        bdo_ready       = '1'           and
                                        decrypt_reg     = '0'
                            ) else '0';

    key_ready   <= key_ready_reg;
    bdi_ready   <= bdi_ready_reg;

    bdo         <=  tag_out                     when (aes_outsel_reg = '1') else        -- Tag
                    (bdi xor ozp_out_node)      when (last_seg = '1') else              -- Last M or C
                    bdo_out0_node;                                                      -- M[i] or C[i]

    bdo_valid        <= d_vld or t_vld;
    bdo_size        <=  CONV_STD_LOGIC_VECTOR(G_TAG_SIZE/8, (G_LBS_BYTES+1))    when (t_vld = '1') else
                        bdi_size_reg;

    msg_auth_done       <= '1' when (  state_reg        = ST_LAST2  and
                                        decrypt_reg     = '1'       and
                                        bdi_valid       = '1'
                                ) else '0';

    msg_auth_valid      <= '1'  when (  state_reg       = ST_LAST2  and
                                        decrypt_reg     = '1'       and
                                        bdi_valid       = '1'       and
                                        bdi = tag_out
                                ) else '0';



end structure;


